import asyncio
import logging

from .exceptions import DeleteMessage

logger = logging.getLogger(__name__)


class LoaferDispatcher:
    def __init__(self, routes, max_jobs=None):
        self.routes = routes
        jobs = max_jobs or len(routes) * 2
        self._semaphore = asyncio.Semaphore(jobs)

    async def dispatch_message(self, message, route):
        logger.debug('dispatching message to route={}'.format(route))
        confirm_message = False
        if not message:
            logger.warning('message will be ignored:\n{!r}\n'.format(message))
            return confirm_message

        with await self._semaphore:
            try:
                confirm_message = await route.deliver(message)
            except DeleteMessage as exc:
                logger.info('explicit message deletion\n{!r}\n'.format(message))
                confirm_message = True
            except asyncio.CancelledError as exc:
                msg = '"{!r}" was cancelled, the message will not be acknowledged:\n{!r}\n'
                logger.warning(msg.format(route.handler, message))
            except Exception as exc:
                logger.exception('{!r}'.format(exc))
                confirm_message = await route.error_handler(type(exc), exc, message)

        return confirm_message

    async def process_route(self, route):
        provider = route.provider
        messages = await provider.fetch_messages()
        for message in messages:
            confirmation = await self.dispatch_message(message, route)
            if confirmation:
                await provider.confirm_message(message)

    async def _dispatch_tasks(self, loop):
        tasks = [self.process_route(route) for route in self.routes]
        await asyncio.wait(tasks, loop=loop)

    async def dispatch_providers(self, loop, forever=True):
        if not forever:
            return await self._dispatch_tasks(loop)

        while True:
            await self._dispatch_tasks(loop)

    def stop_providers(self):
        logger.info('stopping providers')
        for route in self.routes:
            route.provider.stop()



import asyncio
import logging

import aiobotocore
import botocore.exceptions
from cached_property import cached_property

from loafer.exceptions import ProviderError

logger = logging.getLogger(__name__)


class SQSProvider:
    def __init__(self, queue_name, endpoint_url=None, use_ssl=True, options=None, loop=None):
        self.queue_name = queue_name
        self.endpoint_url = endpoint_url
        self.use_ssl = use_ssl
        self._loop = loop or asyncio.get_event_loop()
        self._options = options or {}
        self._queue_url = None

    @cached_property
    def client(self):
        session = aiobotocore.get_session(loop=self._loop)
        return session.create_client('sqs', endpoint_url=self.endpoint_url, use_ssl=self.use_ssl)

    async def get_queue_url(self):
        if not self._queue_url:
            response = await self.client.get_queue_url(QueueName=self.queue_name)
            self._queue_url = response['QueueUrl']
        return self._queue_url

    async def confirm_message(self, message):
        logger.info('confirm message (ack/deletion)')

        receipt = message['ReceiptHandle']
        logger.debug('receipt={!r}'.format(receipt))

        queue_url = await self.get_queue_url()
        return await self.client.delete_message(QueueUrl=queue_url, ReceiptHandle=receipt)

    async def fetch_messages(self):
        try:
            queue_url = await self.get_queue_url()
            logger.debug('fetching messages on {}'.format(queue_url))

            response = await self.client.receive_message(QueueUrl=queue_url, **self._options)
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as exc:
            raise ProviderError('error fetching messages') from exc

        return response.get('Messages', [])

    def stop(self):
        self.client.close()



import logging

logger = logging.getLogger(__name__)


class StringMessageTranslator:

    def translate(self, message):
        logger.debug('{!r} will translate {!r}'.format(type(self).__name__, message))
        return {'content': str(message), 'metadata': {}}



import json

import boto3


def sqs_publish(queue, message):
    _client = boto3.client('sqs')

    if queue.startswith('https://'):
        queue_url = queue
    else:
        response = _client.get_queue_url(QueueName=queue)
        queue_url = response['QueueUrl']

    return _client.send_message(QueueUrl=queue_url, MessageBody=message)


def sns_publish(topic, message):
    _client = boto3.client('sns')
    if topic.startswith('arn:'):
        arn = topic
    else:
        topics = _client.list_topics()
        for topic_data in topics['Topics']:
            if topic_data['TopicArn'].endswith(topic):
                arn = topic_data['TopicArn']
                break

    msg = json.dumps({'default': message})
    return _client.publish(TopicArn=arn, MessageStructure='json', Message=msg)



import json
import logging

logger = logging.getLogger(__name__)


class SQSMessageTranslator:

    def translate(self, message):
        translated = {'content': None, 'metadata': {}}
        try:
            body = message['Body']
        except (KeyError, TypeError):
            logger.error('missing Body key in SQS message. It really came from SQS ?'
                         '\nmessage={!r}'.format(message))
            return translated

        try:
            translated['content'] = json.loads(body)
        except json.decoder.JSONDecodeError as exc:
            logger.error('error={!r}, message={!r}'.format(exc, message))
            return translated

        message.pop('Body')
        translated['metadata'].update(message)
        return translated


class SNSMessageTranslator:

    def translate(self, message):
        translated = {'content': None, 'metadata': {}}
        try:
            body = json.loads(message['Body'])
            message_body = body.pop('Message')
        except (KeyError, TypeError):
            logger.error(
                'Missing Body or Message key in SQS message. It really came from SNS ?'
                '\nmessage={!r}'.format(message))
            return translated

        translated['metadata'].update(message)
        translated['metadata'].pop('Body')

        try:
            translated['content'] = json.loads(message_body)
        except (json.decoder.JSONDecodeError, TypeError) as exc:
            logger.error('error={!r}, message={!r}'.format(exc, message))
            return translated

        translated['metadata'].update(body)
        return translated



import asyncio
import logging

from cached_property import cached_property

from .dispatchers import LoaferDispatcher
from .runners import LoaferRunner

logger = logging.getLogger(__name__)


class LoaferManager:
    def __init__(self, routes, runner=None):
        self.runner = runner or LoaferRunner(on_stop_callback=self.on_loop__stop)
        self.routes = routes

    @cached_property
    def dispatcher(self):
        # TODO: check routes
        return LoaferDispatcher(self.routes)

    def run(self, forever=True):
        loop = self.runner.loop
        self._future = asyncio.gather(
            self.dispatcher.dispatch_providers(loop, forever=forever),
            loop=loop,
        )
        self._future.add_done_callback(self.on_future__errors)
        self.runner.start(self._future, run_forever=forever)

    #
    # Callbacks
    #

    def on_future__errors(self, future):
        exc = future.exception()
        # Unhandled errors crashes the event loop execution
        if isinstance(exc, BaseException):
            logger.critical('fatal error caught: {!r}'.format(exc))
            self.runner.stop()

    def on_loop__stop(self, *args, **kwargs):
        logger.debug('stopping consumers ...')
        self.dispatcher.stop_providers()

        if hasattr(self, '_future'):
            logger.debug('cancel schedulled operations ...')
            self._future.cancel()



import asyncio
from concurrent.futures import ThreadPoolExecutor
import logging
import os
import signal

logger = logging.getLogger(__name__)


class LoaferRunner:
    def __init__(self, loop=None, max_workers=None, on_stop_callback=None):
        self._on_stop_callback = on_stop_callback
        self.loop = loop or asyncio.get_event_loop()

        # XXX: See https://github.com/python/asyncio/issues/258
        # The minimum value depends on the number of cores in the machine
        # See https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor
        self._executor = ThreadPoolExecutor(max_workers)
        self.loop.set_default_executor(self._executor)

    def start(self, future=None, run_forever=True):
        start = 'starting Loafer, pid={}, run_forever={}'
        logger.info(start.format(os.getpid(), run_forever))

        self.loop.add_signal_handler(signal.SIGINT, self.stop)
        self.loop.add_signal_handler(signal.SIGTERM, self.stop)

        try:
            if run_forever:
                self.loop.run_forever()
            else:
                self.loop.run_until_complete(future)
                self.stop()
        finally:
            self.loop.close()

    def stop(self, *args, **kwargs):
        logger.info('stopping Loafer ...')
        if callable(self._on_stop_callback):
            self._on_stop_callback()

        self._executor.shutdown(wait=True)
        if self.loop.is_running():
            self.loop.stop()
